from dune.perftool.pdelab.restriction import restricted_name

from dune.perftool.generation import (get_backend,
                                      temporary_variable,
                                      instruction)
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.options import (get_form_option,
                                   option_switch, get_option)
from dune.perftool.pdelab.geometry import (world_dimension,
                                           local_dimension,
                                           apply_to_global_transformation,
                                           name_facet_jacobian_determinant,
                                           name_element_corners,
                                           component_iname)
from dune.perftool.blockstructured.tools import sub_element_inames
import pymbolic.primitives as prim
from loopy.match import Writes


# TODO warum muss within_inames_is_final=True gesetzt werden?
def compute_corner_diff(first, second, additional_terms=tuple()):
    corners = name_element_corners()
    simplified_names = tuple("cd" + n.split('_')[2] + n.split('_')[3] for n in additional_terms)
    name = "corner_diff_" + "_".join((str(first), str(second)) + simplified_names)
    temporary_variable(name, shape_impl=('fv',), shape=(world_dimension(),))
    diminame = component_iname(context='corner')

    if additional_terms:
        xs_sum = prim.Sum(tuple(prim.Subscript(prim.Variable(x), (prim.Variable(diminame),)) for x in additional_terms))
    else:
        xs_sum = 0

    instruction(expression=prim.Sum((prim.Subscript(prim.Variable(corners), (first, prim.Variable(diminame))),
                                     -1 * prim.Subscript(prim.Variable(corners), (second, prim.Variable(diminame))),
                                     -1 * xs_sum)),
                assignee=prim.Subscript(prim.Variable(name), prim.Variable(diminame)),
                within_inames=frozenset({diminame}),
                within_inames_is_final=True,
                depends_on=frozenset({Writes(corners)} | {Writes(term) for term in additional_terms}))
    return name


# Compute the coefficients for the bilinear geometry transformation
def bilinear_transformation_coefficients():
    dim = world_dimension()
    if dim == 2:
        # Transformation T(x,y) = a * xy + b * x + c * y + d
        c = compute_corner_diff(2, 0)
        b = compute_corner_diff(1, 0)
        a = compute_corner_diff(3, 0, (b, c))
        return [a, b, c]
    elif dim == 3:
        # Transformation T(x,y,z) = a * xyz + b * xy + c * xz + d * yz + e * x + f * y + g * z + h
        g = compute_corner_diff(4, 0)
        f = compute_corner_diff(2, 0)
        e = compute_corner_diff(1, 0)
        d = compute_corner_diff(6, 0, (f, g))
        c = compute_corner_diff(5, 0, (e, g))
        b = compute_corner_diff(3, 0, (e, f))
        a = compute_corner_diff(7, 0, (b, c, d, e, f, g))
        return [a, b, c, d, e, f, g]
    else:
        raise NotImplementedError()


def compute_jacobian(name):
    pymbolic_pos = get_backend("quad_pos", selector=option_switch("blockstructured"))()
    jac_iname = component_iname(context="jac")

    coefficients = bilinear_transformation_coefficients()

    if world_dimension() == 2:
        a, b, c = coefficients

        expr_jac = [None, None]
        expr_jac[0] = prim.Sum((prim.Product((prim.Subscript(pymbolic_pos, (1,)),
                                              prim.Subscript(prim.Variable(a), (prim.Variable(jac_iname),)))),
                                prim.Subscript(prim.Variable(b), (prim.Variable(jac_iname),))))
        expr_jac[1] = prim.Sum((prim.Product((prim.Subscript(pymbolic_pos, (0,)),
                                              prim.Subscript(prim.Variable(a), (prim.Variable(jac_iname),)))),
                                prim.Subscript(prim.Variable(c), (prim.Variable(jac_iname),))))
    elif world_dimension() == 3:
        a, b, c, d, e, f, g = coefficients

        expr_jac = [None, None, None]
        terms = [[b, c, e], [b, d, f], [c, d, g]]

        # j[:][i] = a * qp[k]*qp[l] + v1 * qp[k] + v2 * qp[l] + v3
        # with k, l in {0,1,2} != i and k<l and vj = terms[i][j]
        for i in range(3):
            k, l = sorted(set(range(3)) - {i})
            expr_jac[i] = prim.Sum((prim.Product((prim.Subscript(pymbolic_pos, (k,)), prim.Subscript(pymbolic_pos, (l,)),
                                                  prim.Subscript(prim.Variable(a), (prim.Variable(jac_iname),)))),
                                    prim.Product((prim.Subscript(pymbolic_pos, (k,)),
                                                  prim.Subscript(prim.Variable(terms[i][0]), (prim.Variable(jac_iname),)))),
                                    prim.Product((prim.Subscript(pymbolic_pos, (l,)),
                                                  prim.Subscript(prim.Variable(terms[i][1]), (prim.Variable(jac_iname),)))),
                                    prim.Subscript(prim.Variable(terms[i][2]), (prim.Variable(jac_iname),))
                                    ))
    else:
        raise NotImplementedError()

    for i, expr in enumerate(expr_jac):
        instruction(expression=expr,
                    assignee=prim.Subscript(prim.Variable(name), (prim.Variable(jac_iname), i)),
                    within_inames=frozenset((jac_iname, ) + sub_element_inames() + get_backend(interface="quad_inames")()),
                    depends_on=frozenset({Writes(get_pymbolic_basename(pymbolic_pos))} | {Writes(cd) for cd in coefficients})
                    )


def define_jacobian_matrix(name):
    temporary_variable(name, shape_impl=('fm',), shape=(world_dimension(), world_dimension()))
    compute_jacobian(name)


def name_jacobian_matrix():
    name = "jacobian"
    define_jacobian_matrix(name)
    return name


def compute_determinant(name, matrix):
    dim = world_dimension()
    matrix_entry = [[prim.Subscript(prim.Variable(matrix), (i, j)) for j in range(dim)] for i in range(dim)]
    if dim == 2:
        expr_determinant = prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[1][1])),
                                     -1 * prim.Product((matrix_entry[1][0], matrix_entry[0][1]))))
    elif dim == 3:
        expr_determinant = prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[1][1], matrix_entry[2][2])),
                                     prim.Product((matrix_entry[0][1], matrix_entry[1][2], matrix_entry[2][0])),
                                     prim.Product((matrix_entry[0][2], matrix_entry[1][0], matrix_entry[2][1])),
                                     -1 * prim.Product((matrix_entry[0][2], matrix_entry[1][1], matrix_entry[2][0])),
                                     -1 * prim.Product((matrix_entry[0][0], matrix_entry[1][2], matrix_entry[2][1])),
                                     -1 * prim.Product((matrix_entry[0][1], matrix_entry[1][0], matrix_entry[2][2]))
                                     ))
    else:
        raise NotImplementedError()
    instruction(expression=expr_determinant,
                assignee=prim.Variable(name),
                within_inames=frozenset(sub_element_inames() + get_backend(interface="quad_inames")()),
                depends_on=frozenset({Writes(matrix)})
                )


def define_jacobian_determinant(name):
    temporary_variable(name, shape=())
    jacobian = name_jacobian_matrix()
    compute_determinant(name, jacobian)


def define_jacobian_determinant_inverse(name):
    temporary_variable(name, shape=())
    determinant = name_jacobian_determinant()
    return instruction(expression=prim.Quotient(1., prim.Variable(determinant)),
                       assignee=prim.Variable(name),
                       within_inames=frozenset(sub_element_inames() + get_backend(interface="quad_inames")()),
                       depends_on=frozenset({Writes(determinant)})
                       )


def define_jacobian_determinant_abs(name):
    temporary_variable(name, shape=())
    determinant = name_jacobian_determinant()
    return instruction(expression=prim.Call(prim.Variable("abs"), (prim.Variable(determinant),)),
                       assignee=prim.Variable(name),
                       within_inames=frozenset(sub_element_inames() + get_backend(interface="quad_inames")()),
                       depends_on=frozenset({Writes(determinant)})
                       )


def name_jacobian_determinant():
    name = "detjac"
    define_jacobian_determinant(name)
    return name


def name_jacobian_determinant_inverse():
    name = "detjac_inverse"
    define_jacobian_determinant_inverse(name)
    return name


def name_jacobian_determinant_abs():
    name = "detjac_abs"
    define_jacobian_determinant_abs(name)
    return name


# scale determinant according to the order of the blockstructuring
def pymbolic_jacobian_determinant():
    if get_form_option("constant_transformation_matrix"):
        from dune.perftool.pdelab.geometry import name_jacobian_determinant
        n_det = name_jacobian_determinant()
    else:
        n_det = name_jacobian_determinant_abs()
    return prim.Quotient(prim.Variable(n_det),
                         prim.Power(get_form_option("number_of_blocks"), local_dimension()))


def compute_inverse_transposed(name, det_inv, matrix):
    dim = world_dimension()
    matrix_entry = [[prim.Subscript(prim.Variable(matrix), (i, j)) for j in range(dim)] for i in range(dim)]
    assignee = [[prim.Subscript(prim.Variable(name), (i, j)) for j in range(dim)] for i in range(dim)]
    exprs = [[None for _ in range(dim)] for _ in range(dim)]

    if dim == 2:
        for i in range(2):
            for j in range(2):
                sign = 1. if i == j else -1.
                exprs[i][j] = prim.Product((sign, prim.Variable(det_inv), matrix_entry[1 - i][1 - j]))
    elif dim == 3:
        exprs[0][0] = prim.Product((1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[1][1], matrix_entry[2][2])),
                                              -1 * prim.Product((matrix_entry[1][2], matrix_entry[2][1]))))))
        exprs[1][0] = prim.Product((-1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][1], matrix_entry[2][2])),
                                              -1 * prim.Product((matrix_entry[0][2], matrix_entry[2][1]))))))
        exprs[2][0] = prim.Product((1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][1], matrix_entry[1][2])),
                                              -1 * prim.Product((matrix_entry[0][2], matrix_entry[1][1]))))))

        exprs[0][1] = prim.Product((-1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[1][0], matrix_entry[2][2])),
                                              -1 * prim.Product((matrix_entry[1][2], matrix_entry[2][0]))))))
        exprs[1][1] = prim.Product((1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[2][2])),
                                              -1 * prim.Product((matrix_entry[0][2], matrix_entry[2][0]))))))
        exprs[2][1] = prim.Product((-1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[1][2])),
                                              -1 * prim.Product((matrix_entry[0][2], matrix_entry[1][0]))))))

        exprs[0][2] = prim.Product((1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[1][0], matrix_entry[2][1])),
                                              -1 * prim.Product((matrix_entry[1][1], matrix_entry[2][0]))))))
        exprs[1][2] = prim.Product((-1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[2][1])),
                                              -1 * prim.Product((matrix_entry[0][1], matrix_entry[2][0]))))))
        exprs[2][2] = prim.Product((1., prim.Variable(det_inv),
                                    prim.Sum((prim.Product((matrix_entry[0][0], matrix_entry[1][1])),
                                              -1 * prim.Product((matrix_entry[0][1], matrix_entry[1][0]))))))
    else:
        raise NotImplementedError
    for j in range(dim):
        for i in range(dim):
            instruction(expression=exprs[i][j],
                        assignee=assignee[i][j],
                        within_inames=frozenset(sub_element_inames() + get_backend(interface="quad_inames")()),
                        depends_on=frozenset({Writes(matrix), Writes(det_inv)}))


def define_jacobian_inverse_transposed(name):
    temporary_variable(name, shape_impl=('fm',), shape=(world_dimension(), world_dimension()))
    jacobian = name_jacobian_matrix()
    det_inv = name_jacobian_determinant_inverse()
    compute_inverse_transposed(name, det_inv, jacobian)


def name_jacobian_inverse_transposed(restriction):
    name = restricted_name("jit", restriction)
    define_jacobian_inverse_transposed(name)
    return name


# scale Jacobian according to the order of the blockstructure
def pymbolic_jacobian_inverse(i, j, restriction):
    if get_form_option("constant_transformation_matrix"):
        from dune.perftool.pdelab.geometry import name_constant_jacobian_inverse_transposed
        name_jit = name_constant_jacobian_inverse_transposed(restriction)
    else:
        from dune.perftool.blockstructured.geometry import name_jacobian_inverse_transposed
        name_jit = name_jacobian_inverse_transposed(restriction)
    return prim.Product((get_form_option("number_of_blocks"),
                         prim.Subscript(prim.Variable(name_jit), (j, i))))


# scale determinant according to the order of the blockstructure
def pymbolic_facet_jacobian_determinant():
    return prim.Quotient(prim.Variable(name_facet_jacobian_determinant()),
                         prim.Power(get_form_option("number_of_blocks"), local_dimension()))


# translate a point in the micro element into macro coordinates
def define_point_in_macro(name, point_in_micro):
    dim = local_dimension()
    temporary_variable(name, shape_impl=('fv',), shape=(dim,))

    # point_macro = (point_micro + index_micro) / number_of_blocks
    # where index_micro = tensor index of the micro element
    subelem_inames = sub_element_inames()
    for i in range(dim):
        if isinstance(point_in_micro, prim.Subscript):
            expr = prim.Subscript(point_in_micro.aggregate, point_in_micro.index + (i,))
        else:
            expr = prim.Subscript(point_in_micro, (i,))
        expr = prim.Sum((expr, prim.Variable(subelem_inames[i]),))
        expr = prim.Quotient(expr, get_form_option('number_of_blocks'))
        # TODO relax within inames
        instruction(assignee=prim.Subscript(prim.Variable(name), (i,)),
                    expression=expr,
                    within_inames=frozenset(subelem_inames + get_backend(interface="quad_inames")())
                    )


# TODO add subelem inames if this function gets called
# TODO change input parameter to string
def name_point_in_macro(point_in_micro):
    assert isinstance(point_in_micro, prim.Expression)
    name = get_pymbolic_basename(point_in_micro) + "_macro"
    define_point_in_macro(name, point_in_micro)
    return name


def to_global(local):
    macro = name_point_in_macro(local)
    name = macro + "_global"
    apply_to_global_transformation(name, prim.Variable(macro))
    return prim.Variable(name)
